@@ -6,16 +6,34 @@
 from __future__ import annotations
 
 import inspect
+import json
 import sys
 
 from pathlib import Path
+from types import ModuleType
 from typing import Any
 
 import pytest
 
+import structlog
+
 from structlog import tracebacks
 
 
+class SecretStr(str):  # noqa: SLOT000
+    """
+    Secrets representation as used in Typed Settings or Pydantic.
+    """
+
+    def __repr__(self) -> str:
+        return "*******"
+
+
+@pytest.fixture(autouse=True)
+def _unimport_rich(monkeypatch: pytest.MonkeyPatch) -> None:
+    monkeypatch.setattr(tracebacks, "rich", None)
+
+
 def get_next_lineno() -> int:
     return inspect.currentframe().f_back.f_lineno + 1
 
@@ -47,21 +65,59 @@ def test_safe_str_error():
     ("data", "max_len", "expected"),
     [
         (3, None, "3"),
-        ("spam", None, "spam"),
+        ("spam", None, "'spam'"),
         (b"spam", None, "b'spam'"),
         ("bacon", 3, "'bac'+2"),
         ("bacon", 4, "'baco'+1"),
-        ("bacon", 5, "bacon"),
+        ("bacon", 5, "'bacon'"),
+        (SecretStr("password"), None, "*******"),
+        (["spam", "eggs", "bacon"], 10, "\"['spam', '\"+15"),
     ],
 )
-def test_to_repr(data: Any, max_len: int | None, expected: str):
+def test_to_repr(data: Any, max_len: int | None, expected: str) -> None:
     """
     "to_repr()" returns the repr of an object, trimmed to max_len.
     """
     assert expected == tracebacks.to_repr(data, max_string=max_len)
 
 
-def test_to_repr_error():
+@pytest.mark.parametrize(
+    ("use_rich", "data", "max_len", "expected"),
+    [
+        (True, 3, None, "3"),
+        (True, "spam", None, "'spam'"),
+        (True, b"spam", None, "b'spam'"),
+        (True, "bacon", 3, "'bac'+2"),
+        (True, "bacon", 5, "'bacon'"),
+        (True, SecretStr("password"), None, "*******"),
+        (True, ["spam", "eggs", "bacon"], 4, "['spam', 'eggs', 'baco'+1]"),
+        (False, "bacon", 3, "'bac'+2"),
+        (False, ["spam", "eggs", "bacon"], 4, '"[\'sp"+21'),
+    ],
+)
+def test_to_repr_rich(
+    use_rich: bool,
+    data: Any,
+    max_len: int | None,
+    expected: str,
+    monkeypatch: pytest.MonkeyPatch,
+) -> None:
+    """
+    "to_repr()" uses Rich to get a nice repr if it is installed and if
+    "use_rich" is True.
+    """
+    try:
+        import rich
+    except ImportError:
+        pytest.skip(reason="rich not installed")
+
+    monkeypatch.setattr(tracebacks, "rich", rich)
+    assert expected == tracebacks.to_repr(
+        data, max_string=max_len, use_rich=use_rich
+    )
+
+
+def test_to_repr_error() -> None:
     """
     "to_repr()" does not fail if __repr__() raises an exception.
     """
@@ -97,7 +153,6 @@ def test_simple_exception():
                     filename=__file__,
                     lineno=lineno,
                     name="test_simple_exception",
-                    line="",
                     locals=None,
                 ),
             ],
@@ -130,7 +185,6 @@ def test_raise_hide_cause():
                     filename=__file__,
                     lineno=lineno,
                     name="test_raise_hide_cause",
-                    line="",
                     locals=None,
                 ),
             ],
@@ -164,7 +218,6 @@ def test_raise_with_cause():
                     filename=__file__,
                     lineno=lineno_2,
                     name="test_raise_with_cause",
-                    line="",
                     locals=None,
                 ),
             ],
@@ -179,7 +232,6 @@ def test_raise_with_cause():
                     filename=__file__,
                     lineno=lineno_1,
                     name="test_raise_with_cause",
-                    line="",
                     locals=None,
                 ),
             ],
@@ -208,7 +260,6 @@ def test_raise_with_cause_no_tb():
                     filename=__file__,
                     lineno=lineno,
                     name="test_raise_with_cause_no_tb",
-                    line="",
                     locals=None,
                 ),
             ],
@@ -242,7 +293,6 @@ def test_raise_nested():
                     filename=__file__,
                     lineno=lineno_2,
                     name="test_raise_nested",
-                    line="",
                     locals=None,
                 ),
             ],
@@ -257,7 +307,6 @@ def test_raise_nested():
                     filename=__file__,
                     lineno=lineno_1,
                     name="test_raise_nested",
-                    line="",
                     locals=None,
                 ),
             ],
@@ -287,7 +336,6 @@ def test_raise_no_msg():
                     filename=__file__,
                     lineno=lineno,
                     name="test_raise_no_msg",
-                    line="",
                     locals=None,
                 ),
             ],
@@ -302,7 +350,7 @@ def test_syntax_error():
     try:
         # raises SyntaxError: invalid syntax
         lineno = get_next_lineno()
-        eval("2 +* 2")  # noqa: PGH001
+        eval("2 +* 2")
     except SyntaxError as e:
         trace = tracebacks.extract(type(e), e, e.__traceback__)
 
@@ -323,8 +371,6 @@ def test_syntax_error():
                     filename=__file__,
                     lineno=lineno,
                     name="test_syntax_error",
-                    line="",
-                    locals=None,
                 ),
             ],
         ),
@@ -352,14 +398,12 @@ def test_filename_with_bracket():
                     filename=__file__,
                     lineno=lineno,
                     name="test_filename_with_bracket",
-                    line="",
                     locals=None,
                 ),
                 tracebacks.Frame(
                     filename="<string>",
                     lineno=1,
                     name="<module>",
-                    line="",
                     locals=None,
                 ),
             ],
@@ -388,14 +432,12 @@ def test_filename_not_a_file():
                     filename=__file__,
                     lineno=lineno,
                     name="test_filename_not_a_file",
-                    line="",
                     locals=None,
                 ),
                 tracebacks.Frame(
                     filename=str(Path.cwd().joinpath("string")),
                     lineno=1,
                     name="<module>",
-                    line="",
                     locals=None,
                 ),
             ],
@@ -469,6 +511,12 @@ def test_recursive():
         )
         == frames[0]
     )
+
+    # If we run the tests under Python 3.12 with sysmon enabled, it inserts
+    # frames at the end.
+    if sys.version_info >= (3, 12):
+        frames = [f for f in frames if "coverage" not in f.filename]
+
     # Depending on whether we invoke pytest directly or run tox, either "foo()"
     # or "bar()" is at the end of the stack.
     assert frames[-1] in [
@@ -485,6 +533,46 @@ def test_recursive():
     ]
 
 
+@pytest.mark.parametrize(
+    ("kwargs", "local_variable"),
+    [
+        (
+            {"locals_max_string": None},
+            "x" * (tracebacks.LOCALS_MAX_STRING + 10),
+        ),
+        (
+            {"locals_max_length": None},
+            list(range(tracebacks.LOCALS_MAX_LENGTH + 10)),
+        ),
+    ],
+)
+def test_exception_dict_transformer_locals_max_accept_none_argument(
+    kwargs, local_variable
+):
+    """
+    ExceptionDictTransformer works when either locals_max_string or
+    locals_max_length is set to None.
+    """
+
+    try:
+        my_local = local_variable
+        1 / 0
+    except Exception as e:
+        format_json = tracebacks.ExceptionDictTransformer(
+            show_locals=True, **kwargs
+        )
+        result = format_json((type(e), e, e.__traceback__))
+
+    _ = my_local
+
+    assert len(result) == 1
+    assert len(result[0]["frames"]) == 1
+
+    frame = result[0]["frames"][0]
+
+    assert frame["locals"]["my_local"].strip("'") == str(local_variable)
+
+
 def test_json_traceback():
     """
     Tracebacks are formatted to JSON with all information.
@@ -503,9 +591,7 @@ def test_json_traceback():
             "frames": [
                 {
                     "filename": __file__,
-                    "line": "",
                     "lineno": lineno,
-                    "locals": None,
                     "name": "test_json_traceback",
                 }
             ],
@@ -534,7 +620,6 @@ def test_json_traceback_locals_max_string():
             "frames": [
                 {
                     "filename": __file__,
-                    "line": "",
                     "lineno": lineno,
                     "locals": {
                         "_var": "'spam'+8",
@@ -588,28 +673,176 @@ def test_json_traceback_max_frames(
         if skipped_count:
             assert trace["frames"][skipped_idx] == {
                 "filename": "",
-                "line": "",
                 "lineno": -1,
-                "locals": None,
                 "name": f"Skipped frames: {skipped_count}",
             }
         else:
             assert not any(f["lineno"] == -1 for f in trace["frames"])
 
 
+@pytest.mark.parametrize(
+    ("suppress", "file_no_locals"),
+    [
+        pytest.param((__file__,), __file__, id="file"),
+        pytest.param((json,), json.__file__, id="json"),
+    ],
+)
+def test_json_tracebacks_suppress(
+    suppress: tuple[str | ModuleType, ...],
+    file_no_locals: str,
+    capsys: pytest.CaptureFixture,
+) -> None:
+    """
+    Console and JSON output look as expected
+
+    This means also warnings are logged correctly.
+    """
+    try:
+        try:
+            json.loads(42)  # type: ignore[arg-type]
+        except Exception as e:
+            raise ValueError("error shown to users") from e
+    except ValueError as e:
+        format_json = tracebacks.ExceptionDictTransformer(
+            show_locals=True, suppress=suppress
+        )
+        result = format_json((type(e), e, e.__traceback__))
+        for stack in result:
+            for frame in stack["frames"]:
+                no_locals = frame["filename"] == file_no_locals
+                if no_locals:
+                    assert "locals" not in frame
+                else:
+                    assert "locals" in frame
+
+
+@pytest.mark.parametrize(
+    ("hide_sunder", "hide_dunder", "expected"),
+    [
+        (False, False, {"_spam", "__eggs"}),
+        (True, False, set()),  # Also hides "__eggs", b/c it starts with "_"!
+        (False, True, {"_spam"}),
+        (True, True, set()),
+    ],
+)
+def test_json_tracebacks_skip_sunder_dunder(
+    hide_sunder: bool, hide_dunder: bool, expected: set[str]
+) -> None:
+    """
+    Local variables starting with "_" or "__" can be hidden from the locals
+    dict.
+    """
+
+    def func() -> None:
+        _spam = True
+        __eggs = 3
+        1 / 0
+
+    try:
+        func()
+    except ZeroDivisionError as e:
+        format_json = tracebacks.ExceptionDictTransformer(
+            show_locals=True,
+            locals_hide_sunder=hide_sunder,
+            locals_hide_dunder=hide_dunder,
+        )
+        result = format_json((type(e), e, e.__traceback__))
+        assert expected == set(result[0]["frames"][1]["locals"])
+
+
 @pytest.mark.parametrize(
     "kwargs",
     [
+        {"locals_max_length": -1},
         {"locals_max_string": -1},
         {"max_frames": -1},
         {"max_frames": 0},
         {"max_frames": 1},
+        {"suppress": (json,)},
     ],
 )
-def test_json_traceback_value_error(kwargs):
+def test_json_traceback_value_error(
+    kwargs: dict[str, Any], monkeypatch: pytest.MonkeyPatch
+) -> None:
     """
     Wrong arguments to ExceptionDictTransformer raise a ValueError that
     contains the name of the argument..
     """
+    if "suppress" in kwargs:
+        monkeypatch.setattr(kwargs["suppress"][0], "__file__", None)
     with pytest.raises(ValueError, match=next(iter(kwargs.keys()))):
         tracebacks.ExceptionDictTransformer(**kwargs)
+
+
+class TestLogException:
+    """
+    Higher level integration tests for `Logger.exception()`.
+    """
+
+    @pytest.fixture
+    def cap_logs(self) -> structlog.testing.LogCapture:
+        """
+        Create a LogCapture to be used as processor and fixture for retrieving
+        logs in tests.
+        """
+        return structlog.testing.LogCapture()
+
+    @pytest.fixture
+    def logger(
+        self, cap_logs: structlog.testing.LogCapture
+    ) -> structlog.Logger:
+        """
+        Create a logger with the dict_tracebacks and a LogCapture processor.
+        """
+        old_processors = structlog.get_config()["processors"]
+        structlog.configure([structlog.processors.dict_tracebacks, cap_logs])
+        logger = structlog.get_logger("dict_tracebacks")
+        try:
+            yield logger
+        finally:
+            structlog.configure(processors=old_processors)
+
+    def test_log_explicit_exception(
+        self, logger: structlog.Logger, cap_logs: structlog.testing.LogCapture
+    ) -> None:
+        """
+        The log row contains a traceback when `Logger.exception()` is
+        explicitly called with an exception instance.
+        """
+        try:
+            1 / 0
+        except ZeroDivisionError as e:
+            logger.exception("onoes", exception=e)
+
+        log_row = cap_logs.entries[0]
+
+        assert log_row["exception"][0]["exc_type"] == "ZeroDivisionError"
+
+    def test_log_implicit_exception(
+        self, logger: structlog.Logger, cap_logs: structlog.testing.LogCapture
+    ) -> None:
+        """
+        The log row contains a traceback when `Logger.exception()` is called
+        in an `except` block but without explicitly passing the exception.
+        """
+        try:
+            1 / 0
+        except ZeroDivisionError:
+            logger.exception("onoes")
+
+        log_row = cap_logs.entries[0]
+
+        assert log_row["exception"][0]["exc_type"] == "ZeroDivisionError"
+
+    def test_no_exception(
+        self, logger: structlog.Logger, cap_logs: structlog.testing.LogCapture
+    ) -> None:
+        """
+        `Logger.exception()` should not be called outside an `except` block
+        but this cases is gracefully handled and does not lead to an exception.
+
+        See: https://github.com/hynek/structlog/issues/634
+        """
+        logger.exception("onoes")
+
+        assert [{"event": "onoes", "log_level": "error"}] == cap_logs.entries
